/*
 * bishops:  Gentle Bishops puzzle solver
 * http://www.brouhaha.com/~eric/software/bishops/
 *
 * $Id: bishops.c 9 2004-02-02 02:51:56Z eric $
 * Copyright 1996, 2004 Eric Smith <eric@brouhaha.com>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.  Note that permission is
 * not granted to redistribute this program under the terms of any
 * other version of the General Public License.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111 USA
 */


#include <stdio.h>
#include <string.h>

/*

+----+----+----+----+----+
|  0 |    |  4 |    |  8 |
+----+----+----+----+----+
|    |  2 |    |  6 |    |
+----+----+----+----+----+
|  1 |    |  5 |    |  9 |
+----+----+----+----+----+
|    |  3 |    |  7 |    |
+----+----+----+----+----+

*/


int moves [10] [10] =
{ /*         0  1  2  3  4  5  6  7  8  9 */
  /* 0 */  { 0, 0, 1, 0, 0, 1, 0, 1, 0, 0 },
  /* 1 */  { 0, 0, 1, 1, 1, 0, 0, 0, 0, 0 },
  /* 2 */  { 1, 1, 0, 0, 1, 1, 0, 1, 0, 0 },
  /* 3 */  { 0, 1, 0, 0, 0, 1, 1, 0, 1, 0 },
  /* 4 */  { 0, 1, 1, 0, 0, 0, 1, 0, 0, 1 },
  /* 5 */  { 1, 0, 1, 1, 0, 0, 1, 1, 1, 0 },
  /* 6 */  { 0, 0, 0, 1, 1, 1, 0, 0, 1, 1 },
  /* 7 */  { 1, 0, 1, 0, 0, 1, 0, 0, 0, 1 },
  /* 8 */  { 0, 0, 0, 1, 0, 1, 1, 0, 0, 0 },
  /* 9 */  { 0, 0, 0, 0, 1, 0, 1, 1, 0, 0 }
};


typedef int board [4];

#define BVMAX 200

typedef struct
{
  int pos;
  board b [BVMAX];
} boardvec;

int max_depth = 0;
int solutions = 0;

boardvec shortest_solution;

board start = { 0, 1, 8, 9 };
board goal  = { 8, 9, 0, 1 };

void board_copy (board *dst, board *src)
{
  int i;
  for (i = 0; i < 4; i++)
    (* dst) [i] = (* src) [i];
}

int board_valid (board *b)
{
  return (! (moves [(*b) [0]] [(*b) [2]] ||
	     moves [(*b) [0]] [(*b) [3]] ||
	     moves [(*b) [1]] [(*b) [2]] ||
	     moves [(*b) [1]] [(*b) [3]]));
};

int board_match (board *b1, board *b2)
{
  return (((* b1) [0] == (* b2) [0]) &&
	  ((* b1) [1] == (* b2) [1]) &&
	  ((* b1) [2] == (* b2) [2]) &&
	  ((* b1) [3] == (* b2) [3]));
};

int board_similar (board *b1, board *b2)
{
  /* are both "white" pieces in the same place? */
  if (((* b1) [0] == (* b2) [0]) &&
      ((* b1) [1] == (* b2) [1]))
    {
      /* yes, check "black" pieces */
      if ((* b1) [2] == (* b2) [2])
	return (moves [(*b1) [3]] [(*b2) [3]]);
      if ((* b1) [2] == (* b2) [3])
	return (moves [(*b1) [3]] [(*b2) [2]]);
      if ((* b1) [3] == (* b2) [2])
	return (moves [(*b1) [2]] [(*b2) [3]]);
      if ((* b1) [3] == (* b2) [3])
	return (moves [(*b1) [2]] [(*b2) [2]]);
      return (0);  /* neither black piece matches */
    }
  /* are both "black" pieces in the same place? */
  if (((* b1) [2] == (* b2) [2]) &&
      ((* b1) [3] == (* b2) [3]))
    {
      /* yes, check "white" pieces */
      if ((* b1) [0] == (* b2) [0])
	return (moves [(*b1) [1]] [(*b2) [1]]);
      if ((* b1) [0] == (* b2) [1])
	return (moves [(*b1) [1]] [(*b2) [0]]);
      if ((* b1) [1] == (* b2) [0])
	return (moves [(*b1) [0]] [(*b2) [1]]);
      if ((* b1) [1] == (* b2) [1])
	return (moves [(*b1) [0]] [(*b2) [0]]);
      return (0);  /* neither white piece matches */
    }
  return (0);  /* not even close */
}

int has_loop (boardvec *bv)
{
  int i;

  for (i = 0; i < bv->pos; i++)
    if (board_match (& bv->b [i], & bv->b [bv->pos]))
      return (1);
  return (0);
}

int has_similar (boardvec *bv)
{
  int i;

  for (i = 0; i < bv->pos - 1; i++)
    if (board_similar (& bv->b [i], & bv->b [bv->pos]))
      return (1);
  return (0);
}


int position_empty (board *b, int pos)
{
  return (((* b) [0] != pos) && ((* b) [1] != pos) &&
	  ((* b) [2] != pos) && ((* b) [3] != pos));
};

void print_solution (boardvec *bv)
{
  int i;

  solutions++;

  if (bv->pos < shortest_solution.pos)
    memcpy (& shortest_solution, bv, sizeof (shortest_solution));

  if (bv->pos > shortest_solution.pos)
    return;

  printf ("Solution %d, length %d:\n", solutions, bv->pos);

  for (i = 0; i <= bv->pos; i++)
    {
      printf ("  %d:  %d %d %d %d\n", i,
	      bv->b [i] [0], bv->b [i] [1], bv->b [i] [2], bv->b [i] [3]);
    }
};

void try_moves (boardvec *bv)
{
  int piece;
  int target;
  int tmp;

  if ((bv->pos + 1) >= BVMAX)
    {
      fprintf (stderr, "maximum depth exceeded\n");
      return;
    }

  if ((bv->pos + 1) > shortest_solution.pos)
    return;

  bv->pos++;

  if (bv->pos > max_depth)
    max_depth = bv->pos;

  for (piece = 0; piece < 4; piece++)
    for (target = 0; target <= 9; target++)
      {
	/* is the target position empty? */
	if (! position_empty (& bv->b [bv->pos-1], target))
	  continue;

	/* can piece legally move there? */
	if (! moves [bv->b [bv->pos-1][piece]] [target])
	  continue;

	board_copy (& bv->b [bv->pos], & bv->b [bv->pos-1]);

	/* move the piece */
	bv->b [bv->pos][piece] = target;

	/* any threats? */
	if (! board_valid (& bv->b [bv->pos]))
	  continue;

	/* canonicalize */
	if (bv->b [bv->pos][piece & 2] > bv->b [bv->pos][piece | 1])
	  {
	    tmp = bv->b [bv->pos][piece & 2];
	    bv->b [bv->pos][piece & 2] = bv->b [bv->pos][piece | 1];
	    bv->b [bv->pos][piece | 1] = tmp;
	  }

	/* are we done? */
	if (board_match (& bv->b [bv->pos], & goal))
	  {
	    print_solution (bv);
	    continue;
	  }

	/* have we had this board before? */
	if (has_loop (bv))
	  continue;

	/* have we had a similar board before? */
	if (has_similar (bv))
	  continue;

	try_moves (bv);
      }
  bv->pos--;
}

int main (int argc, char *argv[])
{
  boardvec bv;

  memset (& bv, 0, sizeof (bv));
  bv.pos = 0;

  memset (& shortest_solution, 0, sizeof (shortest_solution));
  shortest_solution.pos = BVMAX;

  board_copy (& bv.b [0], & start);

  try_moves (& bv);

  return (0);
};
